# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2017-2025, by Samuel Williams.

require_relative "container/controller"

# @namespace
module Async
	# @namespace
	module Container
	end
end
